//
// Copyright (c) 1994-1996, David McFadzean
//
// You may distribute under the terms of the Artistic License, 
// as specified in the README file.
//
// $Header: E:/simbiosys/RCS/POPULAT.CPP 1.1 1996/06/26 10:55:00 dbm Exp $

#include <stdlib.h> // for qsort()
#include <stdio.h>
#include "populat.h"
#include "random.h"

////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//	bioPopulation
//
 
bioPopulation::bioPopulation (void)
{
	_size = 0;
	_elite = 1;
	_sorted = FALSE;

	_ptypes = 0;         
	
	_mutationRate = 0.01;
	
	_totalFitness = 0.0;
	_maxFitness = 0.0;
	_minFitness = 1.7E+308;
	_avgFitness = 0.0;
}             

bioPopulation::~bioPopulation (void)
{     
	for (int i = 0; i < _size; i++)
	{
		delete _ptypes[i];
	}

	delete _ptypes;
}             

void
bioPopulation::SetSize (int size)
{
	assert (size > 0);
	assert (_ptypes == 0);
	 
	_size = size;
	_ptypes = new bioPType*[_size];
	
	ASSERT (_ptypes);
	
	for (int i = 0; i < _size; i++)
	{
		_ptypes[i] = GenPType (GenGType());
	}
}             

int
bioPopulation::GetSize (void)
{              
	return _size;
}

void
bioPopulation::SetElite (int elite)
{                                   
	assert (elite > 0);
	
	_elite = elite;
}

int
bioPopulation::GetElite (void)
{
	return _elite;
}

void
bioPopulation::SetMutationRate (double mrate)
{                                   
	assert (mrate > 0.0);
	assert (mrate < 1.0);
	
	_mutationRate = mrate;
}

double
bioPopulation::GetMutationRate (void)
{
	return _mutationRate;
}

bioPType*
bioPopulation::GetPType (int idx)
{
	assert (idx >= 0);
	assert (idx < _size);
	
	return _ptypes[idx];
} 

static int
compFitness (const void* a, const void* b)
{
	bioPType*	p1 = *(bioPType**) a;
	bioPType*	p2 = *(bioPType**) b;
	                              
	ASSERT_VALID (p1);
	ASSERT_VALID (p2);
		                                        
	double		x = p1->GetFitness();
	double		y = p2->GetFitness();
	
	if (x < y)
	{
		return 1;
	}
	else if (x > y)
	{
		return -1;
	}
	else
	{
		return 0;
	}
}

void
bioPopulation::SortByFitness (void)
{	
	qsort (_ptypes, _size, sizeof(_ptypes[0]), compFitness); 

	_totalFitness = 0;
	_maxFitness = 0;
	_minFitness = 1.7E+308;

	for (int i = 0; i < _size; i++)
	{           
		double			fit = _ptypes[i]->GetFitness();
		_totalFitness += fit;
		
		if (fit > _maxFitness)
		{
			_maxFitness = fit;
		} 
		
		if (fit < _minFitness)
		{
			_minFitness = fit;
		}
	}           
	
	_avgFitness = _totalFitness/_size;
	_sorted = TRUE;
	
	Trace ("%f %f %f %f\n", _minFitness, _maxFitness, _totalFitness, _avgFitness);
}

void
bioPopulation::Breed (void)
{	
	if (_sorted == FALSE)
	{
		SortByFitness();
	}

	bioPType**			nextGen = new bioPType*[_size];
   	                                                    
	ASSERT (nextGen);
	
	// Transfer the elite gtypes directly into the next generation.	                                                    
	for (int i = 0; i < _elite; i++)
	{
		nextGen[i] = GenPType(_ptypes[i]->GetGType());
	}
	   
	// Replace all but the elite with children
	for ( ; i < _size; i++)
	{
		bioGType*		mom = _ChooseParent();
		bioGType*		dad = _ChooseParent();
		bioGType*		dna = mom->Crossover (dad);
		
		dna->Mutate (_mutationRate);
		nextGen[i] = GenPType(dna);
	}
	                                                    
	// Extract the elite gtypes so they aren't deleted.
	for (i = 0; i < _elite; i++)
	{
		_ptypes[i]->ExtractGType();
	}

	for (i = 0; i < _size; i++)
	{
		delete _ptypes[i];
	}
	
	delete _ptypes;
	_ptypes = nextGen;

	_sorted = FALSE;
}

bioGType*
bioPopulation::_ChooseParent (void)
{
	double		x = _totalFitness * DRand();
	double 		sum = 0;
	int			i;
	
	for (i = 0; i < _size; i++)
	{
		sum += _ptypes[i]->GetFitness();
		if (sum > x)
		{
			printf ("_ChooseParent: %d of %d\n", i, _size);
			return _ptypes[i]->GetGType();
		}
	}                            
	
	printf ("_ChooseParent: %d of %d\n", i, _size);
	return _ptypes[_size-1]->GetGType();
} 

double
bioPopulation::GetMinFitness (void)
{            
	return _minFitness;
} 

double
bioPopulation::GetMaxFitness (void)
{            
	return _maxFitness;
}

double
bioPopulation::GetAvgFitness (void)
{            
	return _avgFitness;
}


